; $Id: dmx.asm 178 2008-10-10 12:08:54Z pt $
;
;  DMX IO Routines
;
; This is a heavilly modified version of midi.asm from the IIC_MIDI package
; and uart.asm from the IIC_SPEAKJET package
; It will 'currently' only support 352 DMX channels due to the fact that there
; are only 368 bytes of RAM.
; It will respond to a stream of <CHANNEL><VALUE> from I2C. I am not using any
; start or stop characters although this may change if I have problems with timing...
;
; The code expect 3 bytes, <00> <CHANNEL> <LEVEL> 
; I will be enhancing the first byte probably to include blackout with a single command
; and possibly other features
;
; I have also used macros written by Peter Luethi, 08.01.1999, Dietikon, Switzerland 
; to help with getting my head around branching!
; ==========================================================================
;
;  Copyright 2008 Phil Taylor (phil@taylor.org.uk)
;  Copyright 2006 Thorsten Klose (tk@midibox.org)
;  Licensed for personal non-commercial use only.
;  All other rights reserved.
; 
; ==========================================================================

	list	p=16f88
	radix	dec
#include <p16f88.inc>
#include "macros.inc"
#include "globals.inc"

; ==========================================================================

	;; export lables
	global	DMX_Init
	global	DMX_ChannelSet
	global	DMX_SendBuffer
	global	DMX_RxBufferPut
	global 	DMX_Get_Location
	extern	IIC_CMD_Handler
	extern	IIC_Handler
	
	global	DMX_LEDS_Handler
	
	global	DMX_BUF_POINTERA
	global	DMX_BUF_POINTERB	
	global	DMX_BUFFER_TMP
	global	DMX_RX_SIZE		
	global	DMX_COMMAND	
	global	DMX_BANK1
	global	DMX_BANK2
	global 	DMX_BANK3
	global	DMX_BANK4
	extern	IIC_DATA ; Use for TX/RX IIC

; ==========================================================================

shared_ram		udata_shr
DMX_BUF_POINTERA	res	1
DMX_BUF_POINTERB	res	1
DMX_BUFFER_TMP	res	1
DMX_RX_SIZE		res	1
DMX_CHANNEL		res	1 
DMX_TMP_COMMAND		res	1 
DMX_COMMAND		res	1 
LED_TX_CTR		res	1
LED_RX_CTR		res	1
DMX_DELAY_CTR	res	1

gpr0			udata
DMX_BANK1		res	0x4f

gpr1			udata
DMX_BANK2		res	0x4f

gpr2			udata	
DMX_BANK3		res	0x5f

gpr3			udata
DMX_BANK4		res	0x5f


; ==========================================================================
DMX	code
; ==========================================================================

;; --------------------------------------------------------------------------
;;  FUNCTION: DMX_Init
;;  DESCRIPTION: initialises the DMX interface: baudrate and format is set then
;;  Rx and Tx buffer will be emptied
;;  IN:	  -
;;  OUT:  -
;;  USES: -
;; --------------------------------------------------------------------------
DMX_Init

		SWITCHBANK_1_0
		clrf 	PORTB		; Initialize PORTB

		bsf		PORTB,5		; normal status of DMX line is HI (1)
		
		; Setup UART 250K bps
  		movlw	DMX_BAUD_RATE_CONST 			; Baud rate = FOSC/(16(X+1)), 250K bps 8MHz X=1 
		SWITCHBANK_0_1
  		movwf	SPBRG
  		bsf		TXSTA,BRGH	; high speed baud rate generator
  		bcf		TXSTA,SYNC	; select async mode
  		bsf		TXSTA,TXEN	; enable the transmit
  		bsf		TXSTA,TX9D	; Transmit nine data bits
  		bsf		TXSTA,TX9	; one is used for the stop bit
		SWITCHBANK_1_0
  		bsf		RCSTA,SPEN	; serial port enable
  		;bsf		RCSTA,CREN	; continuous receive disable
		SWITCHBANK_0_1
		
		clrf 	DMX_CHANNEL ;; Set DMX_CHANNEL to zero 
		clrf 	DMX_COMMAND ;; Set DMX_CHANNEL to zero 
		clrf 	DMX_TMP_COMMAND ;; Set DMX_CHANNEL to zero 
		clrf 	DMX_RX_SIZE ;; Set DMX_RX_SIZE to zero 
		clrf 	DMX_BUF_POINTERA
		clrf 	DMX_BUF_POINTERB
		
		
DMX_Clear_Registers ; Simple loop to clear buffers (also has various values for testing...)
		btfsc 	DMX_BUF_POINTERB,1 ; Bit 1 is set (2) so finish.
		goto 	DMX_Clear_End
DMX_Clear_Start
		movf	DMX_BUF_POINTERB,W
		call	DMX_ChannelSet
		movf	DMX_BUF_POINTERA,W
		call	DMX_ChannelSet
;		movf	DMX_BUF_POINTERA,W ; These 2 are for testing only
;		swapf		DMX_BUF_POINTERA,W
		movlw	0x00 ; Set channel data to 0
		call	DMX_ChannelSet
		incf	DMX_BUF_POINTERA,1
		btfsc	STATUS,Z
		incf	DMX_BUF_POINTERB,1
		goto 	DMX_Clear_Registers
DMX_Clear_End
		clrf 	DMX_BUF_POINTERA
		clrf 	DMX_BUF_POINTERB
		
		return
		
		
DMX_RxBufferPut
	;; Currently does nothing but light the RX LED!
	;; To enable RX, PIN2 of PORTA needs to be set high, 
	;; This MUST only be done for a short time to allow an RDM receive. (NOT YET DONE!)
	movlw	DMX_LEDS_RXTX_DELAY
	movwf	LED_RX_CTR
	return

;; --------------------------------------------------------------------------
;;  FUNCTION: DMX_ChannelSet
;;  DESCRIPTION: Updates the buffer with DMX Channel/Value information
;; If the DMX_CHANNEL register is 0x00 then store the received data into it.		
;; If it isn't empty, store the received data into the correct memory location.
;; clear the DMX_CHANNEL register and return
;;  IN:     - 1 byte received over IIC http://tutor.al-williams.com/pic-inst.html
;;  OUT:  -
;;  USES: -
;; --------------------------------------------------------------------------
		
DMX_ChannelSet
	    movwf 	DMX_BUFFER_TMP;
	    btfsc 	DMX_RX_SIZE,2
	    goto 	DMX_ClearAll ; Something went 'really' bad as we have received more than 3 bytes! Clear all registers 
		tstf	DMX_RX_SIZE
		bz		DMX_Address_FirstByte ; if DMX_RX_SIZE is 0 start receiving DMX_CHANNEL
	    btfsc 	DMX_RX_SIZE,0
	    goto 	DMX_Address_SecondByte
	    goto 	DMX_Process_Channel
	    return
DMX_Address_FirstByte 
	    incf 	DMX_RX_SIZE,f
	    movf 	DMX_BUFFER_TMP,W
	    movwf 	DMX_TMP_COMMAND
		return
DMX_Address_SecondByte
	    incf 	DMX_RX_SIZE,f
	    movf 	DMX_BUFFER_TMP,W
	    movwf 	DMX_CHANNEL
	    return	
DMX_Process_Channel
	    incf 	DMX_RX_SIZE,f
		; Check the command bit
		; Bit 7 tells us to accept a command
		btfss 	DMX_TMP_COMMAND,7 ; skip next if bit 7 set 
		goto 	DMX_NoCommand
	    movf 	DMX_TMP_COMMAND,W ; move into DMX_COMMAND
		movwf 	DMX_COMMAND ; Bits 1-6 will contain the actual command we are interested in.	
		btfsc 	DMX_COMMAND,6 ; check if Bit6 is high (DMX_COMMAND_SENDx) 
		bcf	RI_N_PORT, RI_N_PIN	; Tell IIC to start sending stream.
		goto DMX_ClearAll; (We don't want to actually process the values)
DMX_NoCommand
		btfsc 	DMX_TMP_COMMAND,0 ; Bit0 is high so use high bank.
		goto 	DMX_Bank4
		movlw	0x50 
		subwf	DMX_CHANNEL,W
		bnc		DMX_Bank1
		movf	DMX_CHANNEL,W
		movlw	0xa0 
		subwf	DMX_CHANNEL,W
		bnc		DMX_Bank2
        goto   	DMX_Bank3  
        return; 
		
DMX_Bank1
	    movf 	DMX_CHANNEL,W
	    addlw 	LOW(DMX_BANK1)  ;get start address of buffer 
	    goto 	DMX_Store
DMX_Bank2
	    movlw 	0x50 ; Size of BANK1
	    subwf 	DMX_CHANNEL,W ;; Remove 80 from DMX Address 
	    addlw 	LOW(DMX_BANK2)  ;get start address of buffer 
	    goto 	DMX_Store
DMX_Bank3
	    movlw 	0xa0 ; Size of BANK1+BANK2
	    subwf 	DMX_CHANNEL,W ; Remove from DMX Address 
	    addlw 	LOW(DMX_BANK3) ;get start address of buffer 
	    goto 	DMX_Store_High
DMX_Bank4	
		movlw	0x60 ; Maximum number of channels in bank4
		subwf	DMX_CHANNEL,W
		bc		DMX_ClearAll
		movf 	DMX_CHANNEL,W
		addlw 	LOW(DMX_BANK4)  ;get start address of buffer  
DMX_Store_High
		SWITCH_FSR_UPPER
DMX_Store	
	    movwf	FSR			;Use FSR to point at current Channel value
	    movf	DMX_BUFFER_TMP,W		;Save data in DMX_BUFFER_TMP to current memory position
	    movwf	INDF		
		SWITCH_FSR_LOWER ; Just in case we were in upper bank!
DMX_ClearAll
	    clrf	DMX_BUFFER_TMP
	    clrf	DMX_CHANNEL
	    clrf	DMX_TMP_COMMAND
	    clrf	DMX_RX_SIZE
		movlw	DMX_LEDS_RXTX_DELAY
		movwf	LED_RX_CTR
	    return

		
; This is the main SendBuffer function, called from MAINLOOP
;
DMX_SendBuffer

		SWITCHBANK_0_1
DMX_SendByte_Poll
		btfss	TXSTA, TRMT
		goto	DMX_SendByte_Poll
		SWITCHBANK_1_0
		BZF	DMX_BUF_POINTERA,DMX_Check_PointerB ; Check PointerB if PointerA is zero.
		goto DMX_Start_Send
DMX_Check_PointerB
		BNZF	DMX_BUF_POINTERB,DMX_Start_Send ; start sending if POINTERB is NOT zero.

		; Send start sequence, <BREAK>DELAY<MAB><DELAY>
		SWITCHBANK_1_0
		bsf		PORTB,5			; Set dmxout hi
		bcf		RCSTA,SPEN		; serial port disable
		SWITCHBANK_0_1
		bcf		TXSTA,TXEN		; Serial transmit disable
		SWITCHBANK_1_0
		bcf		PORTB,5			; Dmxout low, Break

		; Use delay based on current DMX_COMMAND bits 1,2 
		; ANSI Spec: Break: 176uS-352uS MAB 12uS-88uS IB-Gab < 32uS
		; 00 = USITT
		; 01 = Magnulator
		; 10 = Slower
		; 11 = Very Slow
		btfsc 	DMX_COMMAND,2 ; check if Bit2 is high 
		goto 	DMX_Send_Delay_Bit2High
		btfsc 	DMX_COMMAND,1 ; check if Bit1 is high 
		goto 	DMX_Send_Delay_Bit1High
		; This is the USITT Setting 00 (128uS Break, 12uS MAB) 
		movlw	.128 
		call 	DMX_Delay
		bsf		PORTB,5 ; Send DMX Mark-after-break
		movlw	.12 
		call 	DMX_Delay
		goto DMX_Send_Delay_Finished
DMX_Send_Delay_Bit1High
		; This is the Magnulator Setting 01 (88uS Break, 8uS MAB) 
		movlw	.88 
		call 	DMX_Delay
		bsf		PORTB,5 
		movlw	.8 
		call 	DMX_Delay
		goto 	DMX_Send_Delay_Finished
DMX_Send_Delay_Bit2High
		btfsc 	DMX_COMMAND,1 ; check again if bit1 is high 
		goto 	DMX_Send_Delay_BothHigh
		; This is the Slower Setting 10 (176uS Break, 88uS MAB) 
		movlw	.176 
		call 	DMX_Delay
		bsf		PORTB,5 
		movlw	.88 
		call 	DMX_Delay
		goto 	DMX_Send_Delay_Finished
DMX_Send_Delay_BothHigh
		; This is the Very Slow Setting 11 (255uS Break, 128uS MAB) 
		movlw	.255 
		call 	DMX_Delay
		bsf		PORTB,5 
		movlw	.128 
		call 	DMX_Delay

DMX_Send_Delay_Finished
	
		;re-enable serial comms
		SWITCHBANK_1_0
		bsf		RCSTA,SPEN		; serial port enable
		SWITCHBANK_0_1
		bsf		TXSTA,TXEN		; Serial transmit enable
		movlw   0x0				;
		SWITCHBANK_1_0
		movwf	TXREG
		SWITCHBANK_0_1
DMX_SendByte_Poll2
		btfss	TXSTA, TRMT
		goto	DMX_SendByte_Poll2
		SWITCHBANK_1_0
		
DMX_Start_Send
		; Insert short (10uS) delay between channels 
		movlw	.10
		btfsc 	DMX_COMMAND,2 ; If bit 2 is high (slow send) insert short delay before sending channel 
		call 	DMX_Delay		
		btfsc	DMX_BUF_POINTERB,0 ; Bit0 is high so use BANK4
		goto 	DMX_SendBank4
		
		movlw	0x50
		subwf	DMX_BUF_POINTERA,W
		bc		DMX_SendBank1
		movf	DMX_BUF_POINTERA,W
		movlw	0xa0 
		subwf	DMX_BUF_POINTERA,W
		bc		DMX_SendBank2
        goto   	DMX_SendBank3  
		
DMX_SendBank1
	    movf 	DMX_BUF_POINTERA,W
	    addlw 	LOW(DMX_BANK1)  ;get start address of buffer 
	    goto 	DMX_SendByte_Low
DMX_SendBank2
	    movlw 	0x50
	    subwf 	DMX_BUF_POINTERA,W ;; Remove from DMX Address
	    addlw 	LOW(DMX_BANK2)  ;get start address of buffer  
	    goto 	DMX_SendByte_Low
DMX_SendBank3
	    movlw 	0xa0
	    subwf 	DMX_BUF_POINTERA,W ;; Remove from DMX Address
	    addlw 	LOW(DMX_BANK3)  ;get start address of buffer 
	    goto 	DMX_SendByte_High
DMX_SendBank4	
		movlw	0x60 ; Maximum number of channels in bank4
		subwf	DMX_BUF_POINTERA,W
		bc		DMX_SendAll_Finished
	    movf 	DMX_BUF_POINTERA,W
		addlw 	LOW(DMX_BANK4)  ;get start address of buffer  
DMX_SendByte_High
		SWITCH_FSR_UPPER
DMX_SendByte_Low	
	    movwf	FSR			;Use FSR to point at current Channel value
	    movf	INDF,W		
		SWITCH_FSR_LOWER
DMX_SendByte
		;We now have a byte to send in W
		movwf	TXREG
DMX_SendByteFinished
		incf	DMX_BUF_POINTERA,1
		btfsc	STATUS,Z
		incf	DMX_BUF_POINTERB,1
		return	
DMX_SendAll_Finished
		movlw	DMX_LEDS_RXTX_DELAY
		movwf	LED_TX_CTR
		clrf 	DMX_BUF_POINTERA
		clrf 	DMX_BUF_POINTERB

		return

;; --------------------------------------------------------------------
;;  FUNCTION: DMX_LEDS_Handler
;;  DESCRIPTION: this function should be called periodically from a timer
;;  or the mainloop. It decrements the Rx/Tx counters and sets the LEDs 
;;  depending on the counter values
;; --------------------------------------------------------------------------
DMX_LEDS_Handler

DMX_LEDS_Handler_Rx
	    movf	LED_RX_CTR, W
	    bz	DMX_LEDS_Handler_RxOff
DMX_LEDS_Handler_RxOn
	    decf	LED_RX_CTR, F
#if LED_RX_PORT
	    bcf	LED_RX_PORT, LED_RX_PIN
#endif
	    goto	DMX_LEDS_Handler_Rx_End
DMX_LEDS_Handler_RxOff
#if LED_RX_PORT
	    bsf	LED_RX_PORT, LED_RX_PIN
#endif
DMX_LEDS_Handler_Rx_End

DMX_LEDS_Handler_Tx
	    movf	LED_TX_CTR, W
	    bz	DMX_LEDS_Handler_TxOff
DMX_LEDS_Handler_TxOn
	    decf	LED_TX_CTR, F
#if LED_TX_PORT
	    bcf	LED_TX_PORT, LED_TX_PIN
#endif
	goto	DMX_LEDS_Handler_Tx_End
DMX_LEDS_Handler_TxOff
#if LED_TX_PORT
	    bsf	LED_TX_PORT, LED_TX_PIN
#endif
DMX_LEDS_Handler_Tx_End
		return

	
	
DMX_Get_Location
	; This expects the location in W and will return the byte.
	; if DMX_COMMAND bit 0 is set, it will come from the second bank or 0x00
	;clrf	IIC_READ_CTR
	;just to be safe move W into DMX_CHANNEL - It isn't being used at this point.
	movwf	DMX_CHANNEL
	btfsc 	DMX_COMMAND,0 ; Bit0 is high so use high bank.
	goto 	DMX_Get_Bank4
	movlw	0x50 
	subwf	DMX_CHANNEL,W
	bc		DMX_Get_Bank1
	movf	DMX_CHANNEL,W
	movlw	0xa0 
	subwf	DMX_CHANNEL,W
	bc		DMX_Get_Bank2
    goto   	DMX_Get_Bank3  
DMX_Get_Bank1
	movf 	DMX_CHANNEL,W
	addlw 	LOW(DMX_BANK1)  ;get start address of buffer 
	goto 	DMX_Get_Store
DMX_Get_Bank2
	movlw 	0x50 ; Size of BANK1
	subwf 	DMX_CHANNEL,W ;; Remove from DMX Address 
	addlw 	LOW(DMX_BANK2)  ;get start address of buffer 
	goto 	DMX_Get_Store
DMX_Get_Bank3
	movlw 	0xa0 ; Size of BANK1+BANK2
	subwf 	DMX_CHANNEL,W ;; Remove from DMX Address 
	addlw 	LOW(DMX_BANK3) ;get start address of buffer 
	goto 	DMX_Get_Store_High
DMX_Get_Bank4
	movlw	0x60 ; Maximum number of channels in bank4
	subwf	DMX_CHANNEL,W
	bc		DMX_Get_Store_Overload
	movf 	DMX_CHANNEL,W
	addlw 	LOW(DMX_BANK4)  ;get start address of buffer  
DMX_Get_Store_High
	SWITCH_FSR_UPPER
DMX_Get_Store	
	movwf	FSR			;Use FSR to point at current Channel value
	movf	INDF,W		
	SWITCH_FSR_LOWER
	goto 	DMX_Get_Location_End
DMX_Get_Store_Overload
	movlw	0x0 ; Not sure about this.....
DMX_Get_Location_End
	return
		

; Uses Timer2 scaled at 1uS.
; It MAY take longer than the pre-defined time if I2C is being received but
; I would rather wait for it to complete before I send the frame.....
DMX_Delay
	; Setup TIMER2 
	clrf TMR2 ; Set timer2 to zero
	movwf PR2; Move timer amount into timer2 PR2 register 
	movlw	b'00100100' ;Timer2 ticks at 1uS (timer on)
	movwf	T2CON ;setup pre/post scaler and turn on timer2
	bcf	PIR1, TMR2IF		;clear timer2 flag
DMX_Delay_Loop
	btfss	PIR1, SSPIF
	goto	DMX_Delay_Loop_NoIICEvent
	bcf	PIR1, SSPIF	; clear SSP flag
	call	IIC_Handler	; call IIC handler in iic.inc
	SWITCHBANK_0_1
	movf	SSPSTAT, W
	SWITCHBANK_1_0
	andlw	1 << P
	bz	DMX_Delay_Loop
DMX_Delay_Loop_NoIICEvent
	btfss	PIR1, TMR2IF
	goto DMX_Delay_Loop
	bcf	T2CON, TMR2ON ; Turn off timer until we need it again.	
	bcf	PIR1, TMR2IF		;clear timer2 flag
	return

	
#if 0
DMX_Delay
        movwf   DMX_DELAY_CTR      
        btfsc   DMX_DELAY_CTR,0    
        goto    DMX_Delay_1_Loop     
DMX_Delay_1                    
        nop                    
        rrf     DMX_DELAY_CTR,F   
        movlw   3               
        subwf   DMX_DELAY_CTR,W    
        andlw   b'01111111'    
DMX_Delay_1_Loop            
        addlw   -1          
        btfss   STATUS,Z    
        goto    DMX_Delay_1_Loop
        return                
DMX_Delay                
        goto    $+1   
        addlw   -1 

DMX_Delay_1_Loop
        addlw   -1
        btfsc   STATUS,Z
        return 
		; Check IIC while in loop. Will slow loop down a bit.
		; Especially if we receive a byte
		;btfss	PIR1, SSPIF
		;goto	DMX_Delay_1_Loop
		;bcf	PIR1, SSPIF	; clear SSP flag
		;call	IIC_Handler	; call IIC handler in iic.inc
        goto    DMX_Delay_1_Loop
		
#ENDIF
; ==========================================================================
END

	
