; $Id: dmx.asm 178 2008-10-10 12:08:54Z pt $
;
;  DMX IO Routines
;
; This is a heavilly modified version of midi.asm from the IIC_MIDI package
; and uart.asm from the IIC_SPEAKJET package
; It will 'currently' only support 352 DMX channels due to the fact that there
; are only 368 bytes of RAM.
; It will respond to a stream of <CHANNEL><VALUE> from I2C. I am not using any
; start or stop characters although this may change if I have problems with timing...
;
; The code expect 3 bytes, <00> <CHANNEL> <LEVEL> 
; I will be enhancing the first byte probably to include blackout with a single command
; and possibly other features
;
; I have also used macros written by Peter Luethi, 08.01.1999, Dietikon, Switzerland 
; to help with getting my head around branching!
; ==========================================================================
;
;  Copyright 2008 Phil Taylor (phil@taylor.org.uk)
;  Copyright 2006 Thorsten Klose (tk@midibox.org)
;  Licensed for personal non-commercial use only.
;  All other rights reserved.
; 
; ==========================================================================

	list	p=16f88
	radix	dec
#include <p16f88.inc>
#include "macros.inc"
#include "globals.inc"

; ==========================================================================

	;; export lables
	global	DMX_Init
	global	DMX_ChannelSet
	global	DMX_SendBuffer
	global	DMX_RxBufferPut
	global 	DMX_Get_Location
	
	global	DMX_LEDS_Handler
	
	global	DMX_BUF_POINTER
	global	DMX_RX_SIZE		
	global	DMX_PACKAGE
	global	DMX_COMMAND	

	extern	IIC_CMD_Handler
	extern	IIC_Handler
	extern	IIC_RxBufferUsed
	extern	IIC_RxBufferGet
	extern	IIC_DATA ; Use for TX/RX IIC

; ==========================================================================


gpr0			udata
TMP0			res 1
TMP_CHAN		res 1
CRC8			res 1

gpr2			udata	
DMX_BANK2		res	0x60

gpr3			udata
DMX_BANK3		res	0x60

shared_ram		udata_shr
DMX_BUF_POINTER	res	1 ; This is used by the DMX sending routine.
DMX_PACKAGE		res 4
DMX_COMMAND		res	1 
DMX_RX_SIZE		res	1
LED_TX_CTR		res	1
LED_RX_CTR		res	1
TMP_VAL			res	1



; ==========================================================================
DMX	code
; ==========================================================================

;; --------------------------------------------------------------------------
;;  FUNCTION: DMX_Init
;;  DESCRIPTION: initialises the DMX interface: baudrate and format is set then
;;  Rx and Tx buffer will be emptied
;;  IN:	  -
;;  OUT:  -
;;  USES: -
;; --------------------------------------------------------------------------
DMX_Init

	SWITCHBANK_1_0
	clrf 	PORTB		; Initialize PORTB
	bsf		PORTB,5		; normal status of DMX line is HI (1)
	
	; Setup UART 250K bps
	movlw	DMX_BAUD_RATE_CONST 			; Baud rate = FOSC/(16(X+1)), 250K bps 8MHz X=1 
	SWITCHBANK_0_1
	movwf	SPBRG
	bsf		TXSTA,BRGH	; high speed baud rate generator
	bcf		TXSTA,SYNC	; select async mode
	bsf		TXSTA,TXEN	; enable the transmit
	bsf		TXSTA,TX9D	; Transmit nine data bits
	bsf		TXSTA,TX9	; one is used for the stop bit
	SWITCHBANK_1_0
	bsf		RCSTA,SPEN	; serial port enable
	;bsf		RCSTA,CREN	; continuous receive disable
		
	clrf	DMX_PACKAGE+0	
	clrf	DMX_PACKAGE+1	
	clrf	DMX_PACKAGE+2	
    clrf 	DMX_PACKAGE+3
	clrf 	DMX_RX_SIZE 
	clrf 	DMX_COMMAND
	clrf 	DMX_BUF_POINTER

#if 1	
DMX_Clear_Registers ; Simple loop to clear buffers (also has various values for testing...)
	movlw	0xb0 ; Control Change Message
	call	DMX_ChannelSet_Test
	movf	DMX_BUF_POINTER,w
	call	DMX_ChannelSet_Test
;	movf	DMX_BUF_POINTER,w ; These 2 are for testing only
;	swapf		DMX_BUF_POINTER,w
	movlw	0x00 ; Set channel data to 0
	call	DMX_ChannelSet_Test
	movf	CRC8,w ; Terminate with CRC
	call	DMX_ChannelSet_Test
	incf	DMX_BUF_POINTER,f
	btfsc	STATUS,Z
	goto	DMX_Clear_End
	goto 	DMX_Clear_Registers
DMX_Clear_End
	clrf 	DMX_BUF_POINTER


#else	
DMX_Clear_Registers ; Simple loop to clear buffers (also has various values for testing...)
	movf	DMX_BUF_POINTER,w
	addlw	0x42 ; stream (0x40 + 0x02)
	call	DMX_ChannelSet_Test
	movlw	0x00 ; stream (0x40 + 0x02)
	call	DMX_ChannelSet_Test
	movlw	0xff ; stream (0x40 + 0x02)
	call	DMX_ChannelSet_Test
	movf	DMX_PACKAGE+3,w ; Terminate with CRC
	call	DMX_ChannelSet_Test
DMX_Clear_Loop
	movf	DMX_BUF_POINTER,w
	call	DMX_ChannelSet_Test
	incf	DMX_BUF_POINTER,f
	btfsc	STATUS,Z
	goto 	DMX_Clear_End
	goto 	DMX_Clear_Loop
DMX_Clear_End
	clrf 	DMX_BUF_POINTER
#endif

	return
				
DMX_RxBufferPut
	;; Currently does nothing but light the RX LED!
	;; To enable RX, PIN2 of PORTA needs to be set high, 
	;; This MUST only be done for a short time to allow an RDM receive. (NOT YET DONE!)
	movlw	DMX_LEDS_RXTX_DELAY
	movwf	LED_RX_CTR
	return

;; --------------------------------------------------------------------------
;;  FUNCTION: DMX_ChannelSet
;;  DESCRIPTION: Updates the buffer with DMX Channel/Value information
;; If the DMX_CHANNEL register is 0x00 then store the received data into it.		
;; If it isn't empty, store the received data into the correct memory location.
;; clear the DMX_CHANNEL register and return
;;  IN:     - 1 byte received over IIC http://tutor.al-williams.com/pic-inst.html
;;  OUT:  -
;;  USES: -
;; --------------------------------------------------------------------------
		
DMX_ChannelSet
	call	IIC_RxBufferUsed
	bz	DMX_Channel_Finish
	call	IIC_RxBufferGet
DMX_ChannelSet_Test
    btfsc 	DMX_COMMAND,1 			; This is a stream of channels from the remote
	goto	DMX_Recv_Stream
	tstf	DMX_RX_SIZE
	bz		DMX_Bit1Low_Bit0_Low
    btfsc 	DMX_RX_SIZE,1
	goto	DMX_Bit1High
	goto	DMX_Bit1Low_Bit0High 	; Bit 1 is low so bit 0 must be high!
DMX_Bit1Low_Bit0_Low ; First Byte
	; This section will process the first byte which will contain various commands:
	; So far only CC messages (0xb0-0xbf) will be received but this is where any other
	; commands should be added. (MUST LEAVE BIT6 ALONE AS IT IS COMMAND_MODE)
	; We seem to get spurious 0xff at start-up which screws it up!
	clrf	CRC8
	movwf	DMX_PACKAGE+0 			
	xorlw	0xff
	bz		DMX_Channel_Finish		; DMX_PACKAGE+0 is 0xff which is illegal....	
    incf 	DMX_RX_SIZE,f			; Now we have what may be a valid byte
	movf	DMX_PACKAGE+0,w 			
	call DMX_CRC8			
	btfsc	DMX_PACKAGE+0,6
	return							; This it a command so return ready to get the next byte.
	movf	DMX_PACKAGE+0,w			; CRC function would have replaced W
	xorlw	0xb0
	bz		DMX_Channel_Finish		; DMX_PACKAGE+0 is 0xb0	
	movf	DMX_PACKAGE+0,w 		;(command)
	xorlw	0xb1
	bz		DMX_Channel_Finish		; DMX_PACKAGE+0 is 0xb1	
	goto DMX_Error					; didn't match anything so goto Error.
DMX_Bit1Low_Bit0High ; Second Byte
	movwf	DMX_PACKAGE+1			; CHANNEL
	call DMX_CRC8
    incf 	DMX_RX_SIZE,f
	return 							; Nothing more to do.
DMX_Bit1High ; of DMX_RX_SIZE
    btfsc 	DMX_RX_SIZE,0
	goto DMX_Bit1High_Bit0High
DMX_Bit1High_Bit0Low
	movwf	DMX_PACKAGE+2		; VALUE
	call DMX_CRC8
    incf 	DMX_RX_SIZE,f		; increment 
	return
DMX_Bit1High_Bit0High ; This is the 4th byte so check CRC.
DMX_Check_CRC8 
	; Check that CRC matches before doing anything
	clrf	DMX_RX_SIZE			; Make sure DMX_RX_SIZE is empty 
	subwf 	CRC8,f		; CRC8 and received byte MUST match. 
	bnz		DMX_Error 			; Reset package if CRC doesn't match. (currently 0x00 as I can't get crc working!)
	; CRC Matches so check command.
	movf	CRC8,w
	movwf	DMX_PACKAGE+3
	btfss 	DMX_PACKAGE+0,6 	; If Bit6 is high enter command mode.
	goto 	DMX_Process_Channel
	movf	DMX_PACKAGE+0,w 	;(command)	
	movwf 	DMX_COMMAND 		; Bits 1-6 will contain the actual command we are interested in.		
	btfsc 	DMX_COMMAND,2 		; bit 2 is high, send stream back to client (+1 and +2 contain from and to channels)
	goto DMX_Setup_Stream
	btfss 	DMX_COMMAND,1 		; bit 1 is high, expect to receive stream from client (+1 and +2 contain from and to channels)
	goto	DMX_Setup_RX_Stream
	return	; (We don't want to actually process the values so return and wait for more data.)
DMX_Setup_Stream
	; first set DMX_RX_SIZE to difference between from and to channels.
	; This will be the same for both RX and TX streams:
	movlw	0x0;
	addwf	DMX_PACKAGE+2,w
	subwf	DMX_PACKAGE+1,w
	;bcf	RI_N_PORT, RI_N_PIN		; Tell IIC to start sending stream.
	goto DMX_Channel_Finish
DMX_Setup_RX_Stream
	; DMX_COMMAND+1 is first channel
	movf	DMX_PACKAGE+1,w
	movwf	DMX_RX_SIZE		; Set first channel
	; Help I need one more byte!!!!!! use DMX_PACKAGE+3
	movf	DMX_PACKAGE+2,w
	movwf	DMX_PACKAGE+3
	return
DMX_Recv_Stream
	; We are expecting a stream of values (256)
	; Don't change +0 as this contains the bank!
	movwf	TMP_VAL
	movf	DMX_RX_SIZE,w
	call DMX_Get_Location
	movf	TMP_VAL,w
	movwf	INDF
	SWITCH_FSR_LOWER
	incf	DMX_RX_SIZE,f
	movf	DMX_PACKAGE+2,w		; maximum channel
	subwf	DMX_RX_SIZE,w
	btfsc	STATUS,Z
	clrf	DMX_COMMAND			; we must clear DMX_COMMAND if we have done all channels!
	return
DMX_Process_Channel
	IRQ_DISABLE					; disable interrupts during this phase
	movf	DMX_PACKAGE+1,w
	call DMX_Get_Location ; Find memory location
    movf	DMX_PACKAGE+2,w		
    movwf	INDF	
	SWITCH_FSR_LOWER
	IRQ_ENABLE
	goto	DMX_Store_Stream
	return
DMX_Error	; Use this to set error condition and request retransmit (Also used for channels that don't exist)
    clrf	DMX_RX_SIZE ; Package RX 
    clrf 	DMX_PACKAGE+3
	movlw	0xbf
	movwf	DMX_PACKAGE+0	
	;call 	DMX_CRC8			; calculate CRC8 and store it in DMX_PACKAGE+3
	; We should now have a valid return package with CRC in DMX_PACKAGE+3 and a command_type of 0xbf (error)
DMX_Store_Stream
    btfss 	DMX_COMMAND,1 		; This is a stream of channels from the remote so skip next.
	bcf	RI_N_PORT, RI_N_PIN		; Tell IIC to start sending stream.
	movlw	DMX_LEDS_RXTX_DELAY
	movwf	LED_RX_CTR
DMX_Channel_Finish
	return		

	
	
; This is the main SendBuffer function, called from MAINLOOP
;
DMX_SendBuffer

	SWITCHBANK_0_1
DMX_SendByte_Poll
	btfss	TXSTA, TRMT
	goto	DMX_SendByte_Poll
	SWITCHBANK_1_0
	BNZF	DMX_BUF_POINTER,DMX_Start_Send ; continue sending if pointer is larger than 0.

	; Send start sequence, <BREAK>DELAY<MAB><DELAY>
	SWITCHBANK_1_0
	bsf		PORTB,5			; Set dmxout hi
	bcf		RCSTA,SPEN		; serial port disable
	SWITCHBANK_0_1
	bcf		TXSTA,TXEN		; Serial transmit disable
	SWITCHBANK_1_0
	bcf		PORTB,5			; Dmxout low, Break
	; Use delay based on current DMX_COMMAND bits 1,2 
	; ANSI Spec: Break: 176uS-352uS MAB 12uS-88uS IB-Gab < 32uS
	; 00 = USITT
	; 01 = Magnulator
	; 10 = Slower
	; 11 = Very Slow
	btfsc 	DMX_COMMAND,2 ; check if Bit2 is high 
	goto 	DMX_Send_Delay_Bit2High
	btfsc 	DMX_COMMAND,1 ; check if Bit1 is high 
	goto 	DMX_Send_Delay_Bit1High
	; This is the USITT Setting 00 (128uS Break, 12uS MAB) 
	;movlw	.128  	; The original USITT time was 128uS
	movlw	.176  	; Recommended break time for RDM (to allow responder time to react)
	call 	DMX_Delay
	bsf		PORTB,5 ; Send DMX Mark-after-break
	movlw	.12 
	call 	DMX_Delay
	goto DMX_Send_Delay_Finished
DMX_Send_Delay_Bit1High
	; This is the Magnulator Setting 01 (88uS Break, 8uS MAB) 
	movlw	.88 
	call 	DMX_Delay
	bsf		PORTB,5 
	movlw	.8 
	call 	DMX_Delay
	goto 	DMX_Send_Delay_Finished
DMX_Send_Delay_Bit2High
	btfsc 	DMX_COMMAND,1 ; check again if bit1 is high 
	goto 	DMX_Send_Delay_BothHigh
	; This is the Slower Setting 10 (176uS Break, 88uS MAB) 
	movlw	.176 
	call 	DMX_Delay
	bsf		PORTB,5 
	movlw	.88 
	call 	DMX_Delay
	goto 	DMX_Send_Delay_Finished
DMX_Send_Delay_BothHigh
	; This is the Very Slow Setting 11 (255uS Break, 128uS MAB) 
	movlw	.255 
	call 	DMX_Delay
	bsf		PORTB,5 
	movlw	.128 
	call 	DMX_Delay
DMX_Send_Delay_Finished

	;re-enable serial comms
	SWITCHBANK_1_0
	bsf		RCSTA,SPEN		; serial port enable
	SWITCHBANK_0_1
	bsf		TXSTA,TXEN		; Serial transmit enable
	movlw   0x0				;
	SWITCHBANK_1_0
	movwf	TXREG
	SWITCHBANK_0_1
DMX_SendByte_Poll_Check_Status
	btfss	TXSTA, TRMT
	goto	DMX_SendByte_Poll_Check_Status
	SWITCHBANK_1_0
	
DMX_Start_Send
	; Insert short (10uS) delay between channels 
	movlw	.10
	btfsc 	DMX_COMMAND,2 ; If bit 2 is high (slow send) insert short delay before sending channel 
	call 	DMX_Delay		
	IRQ_DISABLE			;We don't want to get an interrupt here really!
	movf	DMX_BUF_POINTER,w
	call DMX_Get_Location ; Find memory location
    movf	INDF,w		  ; Move current memory location to W
	SWITCH_FSR_LOWER
	movwf	TXREG
	IRQ_ENABLE			
DMX_SendByteFinished
	incf	DMX_BUF_POINTER,f
	btfsc	STATUS,Z
	return	
	movlw	DMX_LEDS_RXTX_DELAY
	movwf	LED_TX_CTR
	return

;; --------------------------------------------------------------------
;;  FUNCTION: DMX_LEDS_Handler
;;  DESCRIPTION: this function should be called periodically from a timer
;;  or the mainloop. It decrements the Rx/Tx counters and sets the LEDs 
;;  depending on the counter values
;; --------------------------------------------------------------------------
DMX_LEDS_Handler

DMX_LEDS_Handler_Rx
    movf	LED_RX_CTR, W
    bz	DMX_LEDS_Handler_RxOff
DMX_LEDS_Handler_RxOn
    decf	LED_RX_CTR, F
#if LED_RX_PORT
    bcf	LED_RX_PORT, LED_RX_PIN
#endif
    goto	DMX_LEDS_Handler_Rx_End
DMX_LEDS_Handler_RxOff
#if LED_RX_PORT
    bsf	LED_RX_PORT, LED_RX_PIN
#endif
DMX_LEDS_Handler_Rx_End

DMX_LEDS_Handler_Tx
    movf	LED_TX_CTR, W
    bz		DMX_LEDS_Handler_TxOff
DMX_LEDS_Handler_TxOn
    decf	LED_TX_CTR, F
#if LED_TX_PORT
    bcf	LED_TX_PORT, LED_TX_PIN
#endif
	goto	DMX_LEDS_Handler_Tx_End
DMX_LEDS_Handler_TxOff
#if LED_TX_PORT
	bsf		LED_TX_PORT, LED_TX_PIN
#endif
DMX_LEDS_Handler_Tx_End
	return

	
	
DMX_Get_Location
	; This expects the location in W and will return the byte
	movwf	TMP0
	BRS	0x40, DMX_Get_Bank1 ; Branch if DMX_CHANNEL is smaller than 0x40
	movf	TMP0,W
	BRS	0xA0, DMX_Get_Bank2 ; Branch if DMX_CHANNEL is smaller than 0xA0
    goto   	DMX_Get_Bank3  
DMX_Get_Bank1
	movf 	TMP0,W	
	addlw 	0xa0  ;start address of buffer 
	goto 	DMX_Get_Move
DMX_Get_Bank2
	movlw 	0x40 ; Size of BANK1
	subwf 	TMP0,W ;; Remove from DMX Address 
	addlw 	LOW(DMX_BANK2)  ;get start address of buffer 
	SWITCH_FSR_UPPER
	goto 	DMX_Get_Move
DMX_Get_Bank3
	movlw 	0xa0 ; Size of BANK1+BANK2
	subwf 	TMP0,w ;; Remove from DMX Address 
	addlw 	LOW(DMX_BANK3) ;get start address of buffer 
	SWITCH_FSR_UPPER
DMX_Get_Move	
	movwf	FSR			;Use FSR to point at current Channel value
	;movf	INDF,W		
	return
		

; Uses Timer2 scaled at 1uS.
; It MAY take longer than the pre-defined time if I2C is being received but
; I would rather wait for it to complete before I send the frame.....
DMX_Delay
	; Setup TIMER2 
	clrf TMR2 ; Set timer2 to zero
	movwf PR2; Move timer amount into timer2 PR2 register 
	movlw	b'00100100' ;Timer2 ticks at 1uS (timer on)
	movwf	T2CON ;setup pre/post scaler and turn on timer2
	bcf	PIR1, TMR2IF		;clear timer2 flag
	IRQ_ENABLE				; Make sure interrupts are enabled.....
DMX_Delay_Loop
	btfss	PIR1, TMR2IF
	goto DMX_Delay_Loop
	bcf	T2CON, TMR2ON ; Turn off timer until we need it again.	
	bcf	PIR1, TMR2IF		;clear timer2 flag
	return


; ==========================================================================

; This function calculates a CRC8 based on ta simple CRC implementation
; Pass each byte of the package (0,1,2) through this function.
; It is an implementation that doesn't require a lookup table so will be quite slow.
;
DMX_CRC8
	xorwf	CRC8,f
	clrw
	btfsc	CRC8,0
	xorlw	0x5e
	btfsc	CRC8,1
	xorlw	0xbc
	btfsc	CRC8,2
	xorlw	0x61
	btfsc	CRC8,3
	xorlw	0xc2
	btfsc	CRC8,4
	xorlw	0x9d
	btfsc	CRC8,5
	xorlw	0x23
	btfsc	CRC8,6
	xorlw	0x46
	btfsc	CRC8,7
	xorlw	0x8c
	movwf	CRC8
	return

END

	
