; $Id: main.asm 179 2008-03-08 12:09:54Z tk $
;
;  Source Code of IIC DMX Bridge MBHP_IIC_DMX
;
; ==========================================================================
;
;  Portions Copyright 2008 Phil Taylor (phil@taylor.org.uk)
;  Copyright 2006 Thorsten Klose (tk@midibox.org)
;  Licensed for personal non-commercial use only.
;  All other rights reserved.
; 
; ==========================================================================

	list	p=16f88
	radix	dec
#include <p16f88.inc>
#include "macros.inc"
#include "globals.inc"

;; ==========================================================================
;;  Chip Configuration
;; ==========================================================================

#if INTERNAL_OSC
	__CONFIG    _CONFIG1, _CP_OFF & _CCP1_RB0 & _DEBUG_OFF & _WRT_PROTECT_OFF & _CPD_OFF & _LVP_OFF & _BODEN_ON & _MCLR_OFF & _PWRTE_OFF & _WDT_ON & _INTRC_IO
#else
	__CONFIG    _CONFIG1, _CP_OFF & _CCP1_RB0 & _DEBUG_OFF & _WRT_PROTECT_OFF & _CPD_OFF & _LVP_OFF & _BODEN_ON & _MCLR_OFF & _PWRTE_OFF & _WDT_ON & _HS_OSC
#endif

	__CONFIG    _CONFIG2, _IESO_OFF & _FCMEN_OFF

; ==========================================================================

	;; import lables
	extern	IIC_Init
	extern	IIC_Init_GetAddress
	extern	IIC_Handler
	extern  IIC_DATA ; Use for TX/RX IIC
	extern	DMX_Init
	extern	DMX_SendBuffer
	extern	DMX_RxBufferPut
	extern	DMX_LEDS_Handler
	extern	DMX_ChannelSet

	
; ==========================================================================
shared_ram		udata_shr
W_TMP			res	1
STATUS_TMP		res	1
FSR_TMP			res	1


; ==========================================================================
VECTORS	code
; ==========================================================================

;; ---[ boot vector ]---
BOOT
	org		0x0000
	goto	START
	
;; ---[ irq vector ]---
IRQ
	org	0x0004
	goto	IRQ_Entry
; ==========================================================================
MAIN	code
; ==========================================================================
IRQ_Entry
	movwf	W_TMP
	swapf	STATUS, W
	clrf	STATUS		; switch to bank 0
	movwf	STATUS_TMP
	movf	FSR, W
	movwf	FSR_TMP

IRQ_Handler

IRQ_Handler_IIC
	; Check for IIC IRQ
	btfss	PIR1,SSPIF
	goto	IRQ_Handler_IIC_End
	bcf		PIR1,SSPIF
	call	IIC_Handler ; Handle the IIC byte.
IRQ_Handler_IIC_End
	
IRQ_Handler_Rx
	;; check for UART Receive IRQ
	btfss	PIR1, RCIF
	goto	IRQ_Handler_Rx_End
	bcf	PIR1, RCIF		; clear interrupt flag
	movf	RCREG, W		; put received byte into buffer
	call	DMX_RxBufferPut ; Currently does nothing but hopefully i can use this to RX DMX??
IRQ_Handler_Rx_End

IRQ_Handler_End
	movf	FSR_TMP, W
	movwf	FSR
	swapf	STATUS_TMP, W
	movwf	STATUS
	swapf	W_TMP, F
	swapf	W_TMP, W
	retfie
;
	
; ==========================================================================
START
INIT
	;; initialize ports
	movlw	PORTA_VALUE
	movwf	PORTA
	movlw	PORTB_VALUE
	movwf	PORTB

	SWITCHBANK_0_1
	movlw	TRISA_VALUE
	movwf	TRISA
	movlw	TRISB_VALUE
	movwf	TRISB

	clrf	ANSEL			; disable analog inputs
	movlw	0x07			; disable comparator inputs
	movwf	CMCON

	clrwdt
	
	; Setup TIMER0, prescaler=32
	movlw	b'00000100'
	movwf	OPTION_REG	
	SWITCHBANK_1_0

	SWITCHBANK_0_1
#if INTERNAL_OSC
	movlw	0x70		; select 8 MHz
	movwf	OSCCON
Init_WaitOscStable		; wait until oscillator runs stable
	btfss	OSCCON, IOFS
	goto	Init_WaitOscStable
#endif
	SWITCHBANK_1_0

	movf	PCLATH, W	; store pointer to both condition tables
	iorlw	0x07
	movwf	PCLATH

	;; initialize UART
	call	DMX_Init

	;; initialize IIC
	call	IIC_Init
	;; initialize timer1
	; Setup TIMER1 prescaler=2
	movlw	b'00001101'
	movwf	T1CON


	;; initialize IRQs
	SWITCHBANK_0_1
        bsf		PIE1, RCIE	; enable UART Receive		receive
		bsf		PIE1, SSPIE	; Enable IIC Interrupt
        clrf    PIE2				; clear all other IRQs
		clrf	INTCON
	SWITCHBANK_1_0
        bcf     RCSTA, CREN
        movf    RCREG, W
        bsf     RCSTA, CREN
 		; enable maskable IRQs
        movlw   (1 << GIE) | (1 << PEIE)
        movwf   INTCON
		
	;; turn off power LED (use for error)
#if LED_PWR_PORT
	bsf	LED_PWR_PORT, LED_PWR_PIN
#endif
	bsf	INTCON, TMR0IF ; force timer0 interrupt
	bsf	PIR1, TMR1IF ; force timer1 interrupt

	
MAINLOOP
	;; feed the watchdog
	clrwdt
	
MAINLOOP_Timer0Event
	;; call UART LED handler on timer0 interrupt
	btfss	INTCON, TMR0IF
	goto	MAINLOOP_NoTimer0Event
	bcf	INTCON, TMR0IF ; Clear the interrupt
	;bsf	INTCON, TMR0IE ; Re-enable the interrupt.
#if INTERNAL_OSC
	movlw	0xf9		; every 1.120 mS (@8 MHz)
#else
	movlw	0xf0		; every 1.024 mS (@20 MHz)
#endif
	addwf	TMR0,F
	call	DMX_LEDS_Handler
MAINLOOP_NoTimer0Event

	;; determine device address (allow to change it on-the-fly without device reset)
	call	IIC_Init_GetAddress
	call	DMX_ChannelSet ; Check if we have any data to process.
	
	; Do not send the DMX packet too frequently, this will limit it to approx 60hz
	; This is needed to provide a gap between packets for RDM packets
	; RDM spec is 58mS wait time.
MAINLOOP_Timer1Event
	;; call DMX_SendBuffer handler on timer1 interrupt
	btfss	PIR1, TMR1IF
	goto	MAINLOOP_NoTimer1Event
	call	DMX_SendBuffer
	bcf	PIR1, TMR1IF ; Clear timer1 interrupt

	;movlw	.254 		; 11210hz (updates about 50 times every second)
	movlw	.253 		; 7485hz (updates about 30 times every second)
	;movlw	.227 	; 680hz (updates about 3 times every second - TEST ONLY)
	addwf	TMR1H,f
	;movlw	. 66		; 11210hz (updates about 50 times every second)
	movlw	.100		; 7485hz (updates about 30 times every second)
	;movlw	.91 	; 680hz (updates about 3 times every second - TEST ONLY)
	addwf	TMR1L,f	
	
	
MAINLOOP_NoTimer1Event


	goto	MAINLOOP

;; ==========================================================================
	END
